from pathlib import Path
from typing import List

from fastapi import APIRouter, Depends, HTTPException, status

from id_broker_common.kelvin import kelvin_session, kelvin_session_kwargs, load_kelvin_session_kwargs
from id_broker_common.user import user_getter_by_broker_id
from ucsschool.apis.models import Plugin
from ucsschool.apis.opa import OPAClient, opa_instance
from ucsschool.apis.plugins.auth import oauth2_scheme
from ucsschool.apis.utils import get_logger
from ucsschool.kelvin.client import (
    InvalidRequest,
    SchoolClass,
    SchoolClassResource,
    Session,
    User,
    UserResource,
)

from . import models

# the name of the plugin which will also determine the name of the resource
PLUGIN_NAME: str = "bettermarks"
# the plugins current version. Not used for anything in the ucsschool-apis app
PLUGIN_VERSION: str = "1.0.0"
# the plugins tags:
PLUGIN_TAGS: List[str] = ["bettermarks"]
# the router that will be mounted as the resource under the plugins name:
router: APIRouter = APIRouter()

PLUGIN_SETTINGS_FILE = Path("/etc/ucsschool/apis/id-broker/self-disclosure-api.json")


def setup():
    logger = get_logger()
    logger.info(f"Setup of {PLUGIN_NAME} with version {PLUGIN_VERSION}")
    kelvin_session_kwargs.update(load_kelvin_session_kwargs(PLUGIN_SETTINGS_FILE))


def roles_in_school(user: User, school_name: str) -> List[str]:
    """Helper to get only the roles that are defined for an user within a specific school"""
    roles = set()
    for role_string in user.ucsschool_roles:
        role, context_type, context_name = role_string.split(":")
        if context_type == "school" and school_name == context_name:
            roles.add(role)
    return list(roles)


async def extract_members(schoolclass: SchoolClass, session: Session) -> models.ClassMembers:
    """Helper to get users of a school_class split into teachers and students"""
    logger = get_logger()
    teachers = []
    students = []
    user_name: str
    for user_name in schoolclass.users:
        user = await get_single_user_by_name(user_name, session)
        if not (hasattr(user, "udm_properties") and "brokerID" in user.udm_properties):
            msg = "udm_properties -> brokerID field required (type=value_error.missing)"
            logger.warning(f"user '{user_name}' did not validate: {msg}")
            continue
        roles = roles_in_school(user, schoolclass.school)
        if "student" in roles:
            student = models.Student(
                user_id=user.udm_properties["brokerID"],
                username=user.name,
                firstname=user.firstname,
                lastname=user.lastname,
            )
            students.append(student)
        if "teacher" in roles:
            teacher = models.Teacher(user_id=user.udm_properties["brokerID"], username=user.name)
            teachers.append(teacher)
    return models.ClassMembers(students=students, teachers=teachers)


async def get_single_user_by_broker_id(broker_id: str, session: Session) -> User:
    """Search for a single user by brokerID. Raise HTTPExceptions if not exactly a single user is found."""
    users = await user_getter_by_broker_id(
        broker_id, session
    )  # This will also raise an HTTPException if '*' is in broker_id
    if len(users) > 1:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Did find more than one user with broker-id {broker_id!r}.",
        )
    elif not users:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"No user was found with broker-id {broker_id!r}.",
        )
    return users[0]


async def get_single_user_by_name(name: str, session: Session) -> User:
    """Search for a single user by name."""
    return await UserResource(session=session).get(name=name)


async def schoolclass_getter(school: str, name: str, session: Session) -> SchoolClass:
    """Search for schoolclass. Raise HTTPExceptions if not exactly a single result is returned."""
    try:
        res = [
            obj async for obj in SchoolClassResource(session=session).search(school=school, name=name)
        ]
    except InvalidRequest:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"No class was found for {school}-{name}.",
        )

    if len(res) > 1:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Found more then one class for {school}-{name}.",
        )
    elif not res:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"No class was found for {school}-{name}.",
        )
    return res[0]


@router.get("/users/{user_id}/metadata", response_model=models.User)
async def user_metadata(
    user_id: str,
    token: str = Depends(oauth2_scheme),
    policy_instance: OPAClient = Depends(opa_instance),
    session: Session = Depends(kelvin_session),
):
    user = await get_single_user_by_broker_id(user_id, session)
    await policy_instance.check_policy_true_or_raise(
        "bettermarks_plugin/user", token, request=dict(username=user.name)
    )

    return models.User(
        user_id=user_id,
        username=user.name,
        firstname=user.firstname,
        lastname=user.lastname,
        type=", ".join(user.roles),  # TODO what exactly to use here?
        school_id=user.school,
    )


@router.get("/users/{user_id}/groups", response_model=models.UsersClasses)
async def user_groups(
    user_id: str,
    token: str = Depends(oauth2_scheme),
    policy_instance: OPAClient = Depends(opa_instance),
    session: Session = Depends(kelvin_session),
):
    user: User = await get_single_user_by_broker_id(user_id, session)
    await policy_instance.check_policy_true_or_raise(
        "bettermarks_plugin/user", token, request=dict(username=user.name)
    )

    groups = []
    for school, classes in user.school_classes.items():
        for classname in classes:
            name = f"{school}-{classname}"
            # get student count, which means students only!
            school_class: SchoolClass = await schoolclass_getter(school, classname, session)
            classmembers = await extract_members(school_class, session)
            groups.append(dict(group_id=name, name=name, student_count=len(classmembers.students)))

    return models.UsersClasses(groups=groups)


@router.get("/groups/{group_id}/users", response_model=models.ClassMembers)
async def group_users(
    group_id: str,  # DEMOSCHOOL-Democlass
    token: str = Depends(oauth2_scheme),
    policy_instance: OPAClient = Depends(opa_instance),
    session: Session = Depends(kelvin_session),
):
    school, classname = group_id.split("-")
    school_class: SchoolClass = await schoolclass_getter(school, classname, session)
    classmembers = await extract_members(school_class, session)

    await policy_instance.check_policy_true_or_raise(
        "bettermarks_plugin/teacher",
        token,
        target=classmembers.dict(include={"teachers"}),
    )

    return classmembers


# Accidentally implemented a resource that wasn't part of the story 6608. Oops, I did it again...
#
# @router.get("/schools/{school_id}/groups", response_model=models.Schoolclasses)
# async def school_groups(
#     school_id: str,
#     token: str = Depends(oauth2_scheme),
#     policy_instance: OPAClient = Depends(opa_instance)):
#     session: Session = await kelvin_session()
#     # TODO OPA
#     classes = []
#     async for sc in SchoolClassResource(session=session).search(school=school_id):
#         classes.append(dict(group_id=sc.name,
#                            name=sc.name))
#     return dict(groups=classes)


# This is the object that is referenced in the pyproject.toml as the plugin object:
BettermarksPlugin = Plugin(PLUGIN_NAME, PLUGIN_VERSION, PLUGIN_TAGS, router, setup)
